import json
import os
import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import aioshutil

from ..._utils import (
    _filter_out_none_values_recursively,
    _force_rename,
    _is_uuid,
    _json_dumps,
    _raise_on_duplicate_storage,
    _raise_on_non_existing_storage,
    _unique_key_to_request_id,
)
from ...consts import StorageTypes
from ..file_storage_utils import _delete_request, _update_metadata, _update_request_queue_item

if TYPE_CHECKING:
    from ..memory_storage import MemoryStorage


class RequestQueueClient:
    """TODO: docs."""

    _id: str
    _request_queue_directory: str
    _client: 'MemoryStorage'
    _name: Optional[str]
    _requests: Dict[str, Dict]
    _created_at: datetime
    _accessed_at: datetime
    _modified_at: datetime
    _handled_request_count = 0  # TODO: Does not seem to be implemented in crawelee, always 0
    _pending_request_count = 0

    def __init__(self, *, base_storage_directory: str, client: 'MemoryStorage', id: Optional[str] = None, name: Optional[str] = None) -> None:
        """TODO: docs."""
        self._id = str(uuid.uuid4()) if id is None else id
        self._request_queue_directory = os.path.join(base_storage_directory, name or self._id)
        self._client = client
        self._name = name
        self._requests = {}
        self._created_at = datetime.utcnow()
        self._accessed_at = datetime.utcnow()
        self._modified_at = datetime.utcnow()

    async def get(self) -> Optional[Dict]:
        """TODO: docs."""
        found = _find_or_cache_request_queue_by_possible_id(self._client, self._name or self._id)

        if found:
            await found._update_timestamps(False)
            return found.to_request_queue_info()

        return None

    async def update(self, *, name: Optional[str] = None) -> Dict:
        """TODO: docs."""
        # Check by id
        existing_queue_by_id = _find_or_cache_request_queue_by_possible_id(self._client, self._name or self._id)

        if existing_queue_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.REQUEST_QUEUE, self._id)

        # Skip if no changes
        if name is None:
            return existing_queue_by_id.to_request_queue_info()

        # Check that name is not in use already
        existing_queue_by_name = next(
            (queue for queue in self._client._request_queues_handled if queue._name and queue._name.lower() == name.lower()), None)

        if existing_queue_by_name is not None:
            _raise_on_duplicate_storage(StorageTypes.REQUEST_QUEUE, 'name', name)

        existing_queue_by_id._name = name

        previous_dir = existing_queue_by_id._request_queue_directory

        existing_queue_by_id._request_queue_directory = os.path.join(self._client._request_queues_directory, name)

        await _force_rename(previous_dir, existing_queue_by_id._request_queue_directory)

        # Update timestamps
        await existing_queue_by_id._update_timestamps(True)

        return existing_queue_by_id.to_request_queue_info()

    async def delete(self) -> None:
        """TODO: docs."""
        queue = next((queue for queue in self._client._request_queues_handled if queue._id == self._id), None)

        if queue is not None:
            self._client._request_queues_handled.remove(queue)
            queue._pending_request_count = 0
            queue._requests.clear()

            if os.path.exists(queue._request_queue_directory):
                await aioshutil.rmtree(queue._request_queue_directory)

    async def list_head(self, *, limit: Optional[int] = None) -> Dict:
        """TODO: docs."""
        existing_queue_by_id = _find_or_cache_request_queue_by_possible_id(self._client, self._name or self._id)

        if existing_queue_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.REQUEST_QUEUE, self._id)

        await existing_queue_by_id._update_timestamps(False)

        items: List[Dict] = []

        for request in existing_queue_by_id._requests.values():
            if len(items) == limit:
                break

            if request['orderNo']:
                items.append(request)

        return {
            'limit': limit,
            'hadMultipleClients': False,
            'queueModifiedAt': existing_queue_by_id._modified_at,
            'items': list(map(lambda item: self._json_to_request(item['json']), items)),
        }

    async def add_request(self, request: Dict, *, forefront: Optional[bool] = None) -> Dict:
        """TODO: docs."""
        # TODO: Throw if uniqueKey or url missing from request dict, also do for update_request...
        existing_queue_by_id = _find_or_cache_request_queue_by_possible_id(self._client, self._name or self._id)

        if existing_queue_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.REQUEST_QUEUE, self._id)

        request_model = self._create_internal_request(request, forefront)

        existing_request_with_id = existing_queue_by_id._requests.get(request_model['id'])

        # We already have the request present, so we return information about it
        if existing_request_with_id is not None:
            await existing_queue_by_id._update_timestamps(False)

            return {
                'requestId': existing_request_with_id['id'],
                'wasAlreadyHandled': existing_request_with_id['orderNo'] is None,
                'wasAlreadyPresent': True,
            }

        existing_queue_by_id._requests[request_model['id']] = request_model
        # TODO: Validate the next line logic, seems wrong in crawlee
        existing_queue_by_id._pending_request_count += 0 if request_model['orderNo'] is None else 1
        await existing_queue_by_id._update_timestamps(True)
        await _update_request_queue_item(
            request=request_model,
            request_id=request_model['id'],
            entity_directory=existing_queue_by_id._request_queue_directory,
            persist_storage=self._client._persist_storage,
        )

        return {
            'requestId': request_model['id'],
            # We return wasAlreadyHandled: false even though the request may
            # have been added as handled, because that's how API behaves.
            'wasAlreadyHandled': False,
            'wasAlreadyPresent': False,
        }

    async def get_request(self, request_id: str) -> Optional[Dict]:
        """TODO: docs."""
        existing_queue_by_id = _find_or_cache_request_queue_by_possible_id(self._client, self._name or self._id)

        if existing_queue_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.REQUEST_QUEUE, self._id)

        await existing_queue_by_id._update_timestamps(False)

        request = existing_queue_by_id._requests.get(request_id)
        return self._json_to_request(request['json'] if request is not None else None)

    async def update_request(self, request: Dict, *, forefront: Optional[bool] = None) -> Dict:
        """TODO: docs."""
        existing_queue_by_id = _find_or_cache_request_queue_by_possible_id(self._client, self._name or self._id)

        if existing_queue_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.REQUEST_QUEUE, self._id)

        request_model = self._create_internal_request(request, forefront)

        # First we need to check the existing request to be
        # able to return information about its handled state.

        existing_request = existing_queue_by_id._requests.get(request_model['id'])

        # Undefined means that the request is not present in the queue.
        # We need to insert it, to behave the same as API.
        if existing_request is None:
            return await self.add_request(request, forefront=forefront)

        # When updating the request, we need to make sure that
        # the handled counts are updated correctly in all cases.
        existing_queue_by_id._requests[request_model['id']] = request_model

        pending_count_adjustment = 0
        is_request_handled_state_changing = type(existing_request['orderNo']) != type(request_model['orderNo'])  # noqa
        request_was_handled_before_update = existing_request['orderNo'] is None

        # We add 1 pending request if previous state was handled
        # TODO: Validate the next 2 lines logic, seems wrong in crawlee
        if is_request_handled_state_changing:
            pending_count_adjustment = 1 if request_was_handled_before_update else -1

        existing_queue_by_id._pending_request_count += pending_count_adjustment
        await existing_queue_by_id._update_timestamps(True)
        await _update_request_queue_item(
            request=request_model,
            request_id=request_model['id'],
            entity_directory=existing_queue_by_id._request_queue_directory,
            persist_storage=self._client._persist_storage,
        )

        return {
            'requestId': request_model['id'],
            'wasAlreadyHandled': request_was_handled_before_update,
            'wasAlreadyPresent': True,
        }

    async def delete_request(self, request_id: str) -> None:
        """TODO: docs."""
        existing_queue_by_id = _find_or_cache_request_queue_by_possible_id(self._client, self._name or self._id)

        if existing_queue_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.REQUEST_QUEUE, self._id)

        request = existing_queue_by_id._requests.get(request_id)

        if request:
            del existing_queue_by_id._requests[request_id]
            existing_queue_by_id._pending_request_count -= 0 if request['orderNo'] is None else 1
            await existing_queue_by_id._update_timestamps(True)
            await _delete_request(entity_directory=existing_queue_by_id._request_queue_directory, request_id=request_id)

    def to_request_queue_info(self) -> Dict:
        """TODO: docs."""
        return {
            'accessedAt': self._accessed_at,
            'createdAt': self._created_at,
            'hadMultipleClients': False,
            'handledRequestCount': self._handled_request_count,
            'id': self._id,
            'modifiedAt': self._modified_at,
            'name': self._name,
            'pendingRequestCount': self._pending_request_count,
            'stats': {},
            'totalRequestCount': len(self._requests),
            'userId': '1',
        }

    async def _update_timestamps(self, has_been_modified: bool) -> None:
        """TODO: docs."""
        self._accessed_at = datetime.utcnow()

        if has_been_modified:
            self._modified_at = datetime.utcnow()

        request_queue_info = self.to_request_queue_info()
        await _update_metadata(data=request_queue_info, entity_directory=self._request_queue_directory, write_metadata=self._client._write_metadata)

    def _json_to_request(self, request_json: Optional[str]) -> Optional[dict]:
        if request_json is None:
            return None
        request = json.loads(request_json)
        return _filter_out_none_values_recursively(request)

    def _create_internal_request(self, request: Dict, forefront: Optional[bool]) -> Dict:
        order_no = self._calculate_order_no(request, forefront)
        id = _unique_key_to_request_id(request['uniqueKey'])

        if request.get('id') is not None and request['id'] != id:
            raise ValueError('Request ID does not match its unique_key.')

        json_request = _json_dumps({**request, 'id': id})
        return {
            'id': id,
            'json': json_request,
            'method': request.get('method'),
            'orderNo': order_no,
            'retryCount': request.get('retryCount', 0),
            'uniqueKey': request['uniqueKey'],
            'url': request['url'],
        }

    def _calculate_order_no(self, request: Dict, forefront: Optional[bool]) -> Optional[int]:
        if request.get('handledAt') is not None:
            return None

        timestamp = int(round(datetime.utcnow().timestamp()))

        return -timestamp if forefront else timestamp


def _find_or_cache_request_queue_by_possible_id(client: 'MemoryStorage', entry_name_or_id: str) -> Optional['RequestQueueClient']:
    # First check memory cache
    found = next((queue for queue in client._request_queues_handled
                  if queue._id == entry_name_or_id or (queue._name and queue._name.lower() == entry_name_or_id.lower())), None)

    if found is not None:
        return found

    request_queues_dir = os.path.join(client._request_queues_directory, entry_name_or_id)
    # Check if directory exists
    if not os.access(request_queues_dir, os.F_OK):
        return None

    id: Union[str, None] = None
    name: Union[str, None] = None
    created_at = datetime.utcnow()
    accessed_at = datetime.utcnow()
    modified_at = datetime.utcnow()
    handled_request_count = 0
    pending_request_count = 0
    entries: List[Dict] = []

    # Access the request queue folder
    for entry in os.scandir(request_queues_dir):
        if entry.is_file():
            if entry.name == '__metadata__.json':
                # We have found the queue's metadata file, build out information based on it
                with open(os.path.join(request_queues_dir, entry.name)) as f:
                    metadata = json.load(f)
                id = metadata['id']
                name = metadata['name']
                created_at = datetime.strptime(metadata['createdAt'], '%Y-%m-%dT%H:%M:%S.%fZ')
                accessed_at = datetime.strptime(metadata['accessedAt'], '%Y-%m-%dT%H:%M:%S.%fZ')
                modified_at = datetime.strptime(metadata['modifiedAt'], '%Y-%m-%dT%H:%M:%S.%fZ')
                handled_request_count = metadata['handledRequestCount']
                pending_request_count = metadata['pendingRequestCount']

                continue

            with open(os.path.join(request_queues_dir, entry.name)) as f:
                request = json.load(f)
            entries.append(request)

    if id is None and name is None:
        is_uuid = _is_uuid(entry_name_or_id)

        if is_uuid:
            id = entry_name_or_id
        else:
            name = entry_name_or_id

    new_client = RequestQueueClient(base_storage_directory=client._request_queues_directory, client=client, id=id, name=name)

    # Overwrite properties
    new_client._accessed_at = accessed_at
    new_client._created_at = created_at
    new_client._modified_at = modified_at
    new_client._handled_request_count = handled_request_count
    new_client._pending_request_count = pending_request_count

    for request in entries:
        new_client._requests[request['id']] = request

    client._request_queues_handled.append(new_client)

    return new_client



import asyncio
import base64
import builtins
import contextlib
import functools
import hashlib
import inspect
import io
import json
import mimetypes
import os
import re
import sys
import time
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Callable, Dict, Generic, NoReturn, Optional, Type, TypeVar, Union, cast, overload

import aioshutil
import psutil
from aiofiles import ospath
from aiofiles.os import remove, rename

from apify_client import __version__ as client_version

from ._version import __version__ as sdk_version
from .consts import (
    _BOOL_ENV_VARS_TYPE,
    _DATETIME_ENV_VARS_TYPE,
    _INTEGER_ENV_VARS_TYPE,
    _STRING_ENV_VARS_TYPE,
    BOOL_ENV_VARS,
    DATETIME_ENV_VARS,
    INTEGER_ENV_VARS,
    REQUEST_ID_LENGTH,
    ApifyEnvVars,
    StorageTypes,
)


def _log_system_info() -> None:
    python_version = '.'.join([str(x) for x in sys.version_info[:3]])

    print('System info:')
    print(f'    Apify SDK version: {sdk_version}')
    print(f'    Apify Client version: {client_version}')
    print(f'    OS: {sys.platform}')
    print(f'    Python version: {python_version}')
    if _is_running_in_ipython():
        print('    Running in IPython: True')


DualPropertyType = TypeVar('DualPropertyType')
DualPropertyOwner = TypeVar('DualPropertyOwner')


class dualproperty(Generic[DualPropertyType]):  # noqa: N801
    """Descriptor combining `property` and `classproperty`.

    When accessing the decorated attribute on an instance, it calls the getter with the instance as the first argument,
    and when accessing it on a class, it calls the getter with the class as the first argument.
    """

    def __init__(self, getter: Callable[..., DualPropertyType]) -> None:
        """Initialize the dualproperty.

        Args:
            getter (Callable): The getter of the property.
            It should accept either an instance or a class as its first argument.
        """
        self.getter = getter

    def __get__(self, obj: Optional[DualPropertyOwner], owner: Type[DualPropertyOwner]) -> DualPropertyType:
        """Call the getter with the right object.

        Args:
            obj (Optional[T]): The instance of class T on which the getter will be called
            owner (Type[T]): The class object of class T on which the getter will be called, if obj is None

        Returns:
            The result of the getter.
        """
        return self.getter(obj or owner)


def _maybe_extract_enum_member_value(maybe_enum_member: Any) -> Any:
    if isinstance(maybe_enum_member, Enum):
        return maybe_enum_member.value
    return maybe_enum_member


@overload
def _fetch_and_parse_env_var(env_var: _BOOL_ENV_VARS_TYPE) -> Optional[bool]:  # noqa: U100
    ...


@overload
def _fetch_and_parse_env_var(env_var: _BOOL_ENV_VARS_TYPE, default: bool) -> bool:  # noqa: U100
    ...


@overload
def _fetch_and_parse_env_var(env_var: _DATETIME_ENV_VARS_TYPE) -> Optional[Union[datetime, str]]:  # noqa: U100
    ...


@overload
def _fetch_and_parse_env_var(env_var: _DATETIME_ENV_VARS_TYPE, default: datetime) -> Union[datetime, str]:  # noqa: U100
    ...


@overload
def _fetch_and_parse_env_var(env_var: _INTEGER_ENV_VARS_TYPE) -> Optional[int]:  # noqa: U100
    ...


@overload
def _fetch_and_parse_env_var(env_var: _INTEGER_ENV_VARS_TYPE, default: int) -> int:  # noqa: U100
    ...


@overload
def _fetch_and_parse_env_var(env_var: _STRING_ENV_VARS_TYPE, default: str) -> str:  # noqa: U100
    ...


@overload
def _fetch_and_parse_env_var(env_var: _STRING_ENV_VARS_TYPE) -> Optional[str]:  # noqa: U100
    ...


@overload
def _fetch_and_parse_env_var(env_var: ApifyEnvVars) -> Optional[Any]:  # noqa: U100
    ...


def _fetch_and_parse_env_var(env_var: Any, default: Any = None) -> Any:
    env_var_name = str(_maybe_extract_enum_member_value(env_var))

    val = os.getenv(env_var_name)
    if not val:
        return default

    if env_var in BOOL_ENV_VARS:
        return _maybe_parse_bool(val)
    if env_var in INTEGER_ENV_VARS:
        res = _maybe_parse_int(val)
        if res is None:
            return default
        return res
    if env_var in DATETIME_ENV_VARS:
        return _maybe_parse_datetime(val)
    return val


def _get_cpu_usage_percent() -> float:
    return psutil.cpu_percent()


def _get_memory_usage_bytes() -> int:
    current_process = psutil.Process(os.getpid())
    mem = int(current_process.memory_info().rss or 0)
    for child in current_process.children(recursive=True):
        try:
            mem += int(child.memory_info().rss or 0)
        except psutil.NoSuchProcess:
            pass
    return mem


def _maybe_parse_bool(val: Optional[str]) -> bool:
    if val == 'true' or val == 'True' or val == '1':
        return True
    return False


def _maybe_parse_datetime(val: str) -> Union[datetime, str]:
    try:
        return datetime.strptime(val, '%Y-%m-%dT%H:%M:%S.%fZ').replace(tzinfo=timezone.utc)
    except ValueError:
        return val


def _maybe_parse_int(val: str) -> Optional[int]:
    try:
        return int(val)
    except ValueError:
        return None


async def _run_func_at_interval_async(func: Callable, interval_secs: float) -> None:
    started_at = time.time()
    while True:
        elapsed_secs = time.time() - started_at
        sleep_for_secs = interval_secs - (elapsed_secs % interval_secs)
        await asyncio.sleep(sleep_for_secs)

        res = func()
        if inspect.isawaitable(res):
            await res


async def _force_remove(filename: str) -> None:
    """JS-like rm(filename, { force: true })."""
    with contextlib.suppress(FileNotFoundError):
        await remove(filename)


def _json_serializer(obj: Any) -> str:  # TODO: Decide how to parse/dump/handle datetimes!
    if isinstance(obj, (datetime)):
        return obj.isoformat(timespec='milliseconds') + 'Z'
    else:
        return str(obj)


def _filter_out_none_values_recursively(dictionary: Dict) -> Dict:
    """Return copy of the dictionary, recursively omitting all keys for which values are None."""
    return cast(dict, _filter_out_none_values_recursively_internal(dictionary))


# Unfortunately, it's necessary to have an internal function for the correct result typing, without having to create complicated overloads
def _filter_out_none_values_recursively_internal(dictionary: Dict, remove_empty_dicts: Optional[bool] = None) -> Optional[Dict]:
    result = {}
    for k, v in dictionary.items():
        if isinstance(v, dict):
            v = _filter_out_none_values_recursively_internal(v, remove_empty_dicts is True or remove_empty_dicts is None)
        if v is not None:
            result[k] = v
    if not result and remove_empty_dicts:
        return None
    return result


def _json_dumps(obj: Any) -> str:
    """Dump JSON to a string with the correct settings and serializer."""
    return json.dumps(obj, ensure_ascii=False, indent=2, default=_json_serializer)


uuid_regex = re.compile('[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}', re.I)


def _is_uuid(string: str) -> bool:
    """Test whether the given string matches UUID format."""
    return bool(uuid_regex.match(string))


def _raise_on_non_existing_storage(client_type: StorageTypes, id: str) -> NoReturn:
    raise ValueError(f'{client_type} with id: {id} does not exist.')


def _raise_on_duplicate_storage(client_type: StorageTypes, key_name: str, value: str) -> NoReturn:
    raise ValueError(f'{client_type} with {key_name}: {value} already exists.')


def _guess_file_extension(content_type: str) -> Optional[str]:
    """Guess the file extension based on content type."""
    # e.g. mimetypes.guess_extension('application/json ') does not work...
    actual_content_type = content_type.split(';')[0].strip()
    ext = mimetypes.guess_extension(actual_content_type)
    # Remove the leading dot if extension successfully parsed
    return ext[1:] if ext is not None else ext


def _is_content_type_json(content_type: str) -> bool:
    return bool(re.search(r'^application/json', content_type, flags=re.IGNORECASE))


def _is_content_type_xml(content_type: str) -> bool:
    return bool(re.search(r'^application/.*xml$', content_type, flags=re.IGNORECASE))


def _is_content_type_text(content_type: str) -> bool:
    return bool(re.search(r'^text/', content_type, flags=re.IGNORECASE))


def _is_file_or_bytes(value: Any) -> bool:
    # The check for IOBase is not ideal, it would be better to use duck typing,
    # but then the check would be super complex, judging from how the 'requests' library does it.
    # This way should be good enough for the vast majority of use cases, if it causes issues, we can improve it later.
    return isinstance(value, (bytes, bytearray, io.IOBase))


def _maybe_parse_body(body: bytes, content_type: str) -> Any:  # TODO: Improve return type
    try:
        if _is_content_type_json(content_type):
            return json.loads(body)  # Returns any
        elif _is_content_type_xml(content_type) or _is_content_type_text(content_type):
            return body.decode('utf-8')  # TODO: Check if utf-8 can be assumed
    except ValueError as err:
        print('_maybe_parse_body error', err)
    return body


def _unique_key_to_request_id(unique_key: str) -> str:
    """Generate request ID based on unique key in a deterministic way."""
    id = re.sub(r'(\+|\/|=)', '', base64.b64encode(hashlib.sha256(unique_key.encode('utf-8')).digest()).decode('utf-8'))

    return id[:REQUEST_ID_LENGTH] if len(id) > REQUEST_ID_LENGTH else id


async def _force_rename(src_dir: str, dst_dir: str) -> None:
    """Rename a directory. Checks for existence of soruce directory and removes destination directory if it exists."""
    # Make sure source directory exists
    if await ospath.exists(src_dir):
        # Remove destination directory if it exists
        if await ospath.exists(dst_dir):
            await aioshutil.rmtree(dst_dir, ignore_errors=True)
        await rename(src_dir, dst_dir)

ImplementationType = TypeVar('ImplementationType', bound=Callable)
MetadataType = TypeVar('MetadataType', bound=Callable)


def _wrap_internal(implementation: ImplementationType, metadata_source: MetadataType) -> MetadataType:
    @functools.wraps(metadata_source)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return implementation(*args, **kwargs)

    return cast(MetadataType, wrapper)


def _is_running_in_ipython() -> bool:
    return getattr(builtins, '__IPYTHON__', False)



import secrets
from collections import OrderedDict
from typing import Generic, Optional
from typing import OrderedDict as OrderedDictType
from typing import TypeVar


def _crypto_random_object_id(length: int = 17) -> str:
    """Python reimplementation of cryptoRandomObjectId from `@apify/utilities`."""
    chars = 'abcdefghijklmnopqrstuvwxyzABCEDFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(secrets.choice(chars) for _ in range(length))


T = TypeVar('T')


# TODO: Check out possible implementations of the LRUCache in Python
# e.g. https://github.com/jlhutch/pylru or https://stackoverflow.com/questions/19775685/how-to-correctly-implement-the-mapping-protocol-in-python
class LRUCache(Generic[T]):
    """Attempt to reimplement LRUCache from `@apify/datastructures` using `OrderedDict`."""

    _cache: OrderedDictType[str, T]

    _max_length: int

    def __init__(self, max_length: int) -> None:
        """Crete a LRUCache with a specific max_length."""
        self._cache = OrderedDict()
        self._max_length = max_length

    def get(self, key: str) -> Optional[T]:
        """Get an item from the cache. Move it to the end if present."""
        val = self._cache.get(key)
        if val is not None:
            self._cache.move_to_end(key)
        return val

    def add(self, key: str, value: T) -> bool:
        """Add an item to the cache. Remove least used item if max_length exceeded."""
        if key in self._cache:
            return False
        self._cache[key] = value
        if len(self._cache) > self._max_length:
            self._cache.popitem(last=False)
        return True

    def remove(self, key: str) -> Optional[T]:
        """Remove an item from the cache."""
        return self._cache.pop(key)

    def clear(self) -> None:
        """Clear the cache."""
        self._cache.clear()

    def __len__(self) -> int:
        """Get the number of items in the cache."""
        return len(self._cache)



import pathlib

from setuptools import find_packages, setup

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

version_file = (here / 'src/apify/_version.py').read_text(encoding='utf-8')
version = None
for line in version_file.splitlines():
    if line.startswith('__version__'):
        delim = '"' if '"' in line else "'"
        version = line.split(delim)[1]
        break
else:
    raise RuntimeError('Unable to find version string.')

setup(
    name='apify',
    version=version,

    author='Apify Technologies s.r.o.',
    author_email='support@apify.com',
    url='https://github.com/apify/apify-sdk-python',
    project_urls={
        'Documentation': 'https://docs.apify.com/apify-sdk-python',
        'Source': 'https://github.com/apify/apify-sdk-python',
        'Issue tracker': 'https://github.com/apify/apify-sdk-python/issues',
        'Apify Homepage': 'https://apify.com',
    },
    license='Apache Software License',
    license_files=['LICENSE'],

    description='Apify SDK for Python',
    long_description=long_description,
    long_description_content_type='text/markdown',

    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
    ],
    keywords='apify, sdk, actor, scraping, automation',

    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    package_data={'apify': ['py.typed']},
    python_requires='>=3.8',
    install_requires=[
        'apify-client ~= 0.7.0b30',
        'httpx ~= 0.23.0',
        'psutil ~= 5.9.4',
        'pydantic ~= 1.10.2',
        'pyee ~= 9.0.4',
        'typing-extensions ~= 4.4.0',
        'websockets ~= 10.4',
        'aiofiles ~= 22.1.0',
        'aioshutil ~= 1.2',
    ],
    extras_require={
        'dev': [
            'autopep8 ~= 2.0.0',
            'flake8 ~= 5.0.4',
            'flake8-bugbear ~= 22.10.27',
            'flake8-commas ~= 2.1.0',
            'flake8-docstrings ~= 1.6.0',
            'flake8-isort ~= 5.0.3',
            'flake8-quotes ~= 3.3.1',
            'flake8-unused-arguments ~= 0.0.12',
            'isort ~= 5.10.1',
            'mypy ~= 0.991',
            'pep8-naming ~= 0.13.2',
            'pre-commit ~= 2.20.0',
            'pytest ~= 7.2.0',
            'pytest-asyncio ~= 0.20.3',
            'pytest-randomly ~= 3.12.0',
            'respx ~= 0.20.1',
            'sphinx ~= 5.3.0',
            'sphinx-autodoc-typehints ~= 1.19.5',
            'sphinx-markdown-builder == 0.5.4',  # pinned to 0.5.4, because 0.5.5 has a formatting bug
            'types-aiofiles ~= 22.1.0.4',
            'types-psutil ~= 5.9.5.5',
            'types-setuptools ~= 65.6.0.1',
        ],
    },
)



from typing import Any, Dict, List, Union

# Type for representing json-serializable values
# It's close enough to the real thing supported by json.parse, and the best we can do until mypy supports recursive types
# It was suggested in a discussion with (and approved by) Guido van Rossum, so I'd consider it correct enough
JSONSerializable = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]
